
############################################################
#                                                          #
#                          Libraries                       #
#                                                          #
############################################################
# 23-05-2023 PM2,5 de 2022
# Substitue all NA's in df_obs_sub by gbm predictions based on ok_total.

# Needs packages:
library(urbanR)
library(gbm)
library(lubridate)
library(dplyr) #per treballar amb cbinds i esas cosas
#library(caret) # an aggregator package for performing many machine learning models

############################################################
#                                                          #
#                        Initiall data                     #
#                                                          #
############################################################
pollutant       <- "pm2p5"
years           <- c(2020, 2021, 2022)
save_result     <-"/esarchive/scratch/mhajji/Meriem/Gene/results_regional_"
file_format     <-"gene"
time_resolution <-"daily"
compute_in_UTM  <- "TRUE"
crs_latlon      <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"
crs_utm         <- "+proj=utm +zone=31 +ellps=intl +units=m +no_defs"

############################################################
#                                                          #
#                 Load Local Stations                      #
#                                                          #
############################################################
#We will pick the latest data available
local_stations     <- readRDS(paste0(save_result,years[length(years)],"/results_gene_",pollutant, "/","local_stations_",file_format, "_",pollutant,".rds"))

############################################################
#                                                          #
#                  Load All observations                   #
#                                                          #
############################################################
df_obs_years <- NULL
for(read_year in years){
  df_obs_all        <- readRDS(file = paste0(save_result,read_year,"/results_gene_",pollutant, "/","df_obs_",file_format,"_",pollutant,".rds"))
  
  start_date  <- as.Date(paste(read_year, "-01-01", sep = ""))
  end_date    <- as.Date(paste(read_year, "-12-30", sep = ""))
  
  dates      <- seq(start_date, end_date, by = "day")
  df_obs_sub <- subset(df_obs_all, date>=start_date & date <= end_date)
  
  
  # Comprovation we don't have duplications in obs
  if(any(duplicated(df_obs_sub$date))){
    print(paste0("Warning, there is duplicated observations on date :", df_obs_sub$date[duplicated(df_obs_sub$date)]))
    df_obs_sub <- distinct(df_obs_sub, date, .keep_all = TRUE)
    print("The df_obs_sub (alias=observations) have been modified to have the same length as the model")
  }
  
  if (length(df_obs_sub$date) != length(dates)){
    print("WARNING: There are missing hours/dates in observations. And the code is not developed for daily data yet!!")
  }
  
  df_obs_years <- rbind.fill(df_obs_years, df_obs_sub)
}

names_local_stations <- as.list(intersect(local_stations@data$ref_obs, colnames(df_all_obs)))

############################################################
#                                                          #
#            Load raw model interpolated to AQMS           #
#                                                          #
############################################################
df_all_model_raw_at_AQMS <- NULL
for(read_year in years){
  
  model_raw_at_AQMS_df     <- readRDS(paste0(save_result,read_year,"/results_gene_",pollutant, "/","model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  df_all_model_raw_at_AQMS <- rbind.fill(df_all_model_raw_at_AQMS, model_raw_at_AQMS_df)
}

############################################################
#                                                          #
#   AQMS interpolated to AQMS as covariate for GBM         #
#                                                          #
############################################################

# We pick up the observations configured in config file:
obs_filled      <- df_all_obs 
observations    <- subset(df_all_obs, select = -date) #we remove dates from observations data frame
OK_other_obs_df <- NULL
# To fill the gaps of observations in each station, so we need a loop to repeat the fill process in each station:
for (s in 1:length(names_local_stations)){
  station <-names_local_stations[[s]]
  # station <- local_stations$ref_obs[1]
  # Data frame amb dos variables date(fecha + hora) i Tests_results/Obs_Test/Obs_Train (NA)
  Test_results_station  <-  data.frame(date=df_all_obs$date , Test_results = NA) 
  Obs_Test_station      <-  data.frame(date=df_all_obs$date , Obs_Test = NA)
  Obs_Train_station     <-  data.frame(date=df_all_obs$date , Obs_Train = NA)
  
  
 # Arregment of AQMS interpolated to AQMS
 # station is the code station that we hidded to interpolate to it, first
  col_station           <- which(station==colnames(df_all_obs))
  stations_to_interp    <- df_all_obs[,-col_station]
  
  
    #date <- df_obs_sub$date[1]
  #OK at station  
    # Build spatialpointdataframe from date
    # Build variogram
    # OK at station
    # Save results 
  c <- 0
  for(j in 1:length(df_all_obs$date)){
    #print(paste0("Date: ", stations_to_interp$date[j]))
    #We take the data day by day /hour by hour
    ok_obs    <- stations_to_interp[j,]
    ok_obs    <- subset(ok_obs,select=-date)
    df_ok_obs <- data.frame(ref_obs = colnames(ok_obs), pollutant = as.numeric(ok_obs))
    
    #coordinates where to interpolate:
    df_hidded_AQMS                      <- local_stations[ which(local_stations$ref_obs == station) , ]
    # df_hidded_AQMS                   <- data.frame(lat=hidded_AQMS@coords[,2],lon=hidded_AQMS@coords[,1],code=hidded_AQMS$ref_obs)
    # coordinates(df_hidded_AQMS)      <- ~lon+lat
    # sp::proj4string(df_hidded_AQMS)  <- crs_latlon
    
    if (compute_in_UTM) {
      df_hidded_AQMS <- spTransform(df_hidded_AQMS, sp::CRS(crs_utm))
    }
    
    #    1.2 Compute Variogram
    #with other AQMS we want a SPDF that has the ref of the station and the coordinates of each, with the value of the pollution in that day/hour
    sp_other_ok_obs <- local_stations[ -which(local_stations$ref_obs == station) , ]
    df_other_obs <- as.data.frame(sp_other_ok_obs)
    df_obs_data  <- merge(df_other_obs[, -which(names(df_other_obs) == "pollutant")], df_ok_obs, by = "ref_obs")
    
    if(!sum(is.na(df_obs_data$pollutant))>=length(df_obs_data$pollutant)){
      df_obs_data                   <- na.omit(df_obs_data)
      coordinates(df_obs_data)      <- ~lon+lat
      sp::proj4string(df_obs_data)  <- crs_latlon
      
      if (compute_in_UTM) {
        df_obs_data <- spTransform(df_obs_data, sp::CRS(crs_utm))
      }
      
      pollutant.fit_ok          <- automap::autofitVariogram(formula=pollutant~1, input_data=df_obs_data,model="Ste")
      
      ok_gstat <- gstat(formula=pollutant~1, data = df_obs_data, model = pollutant.fit_ok$var_model, nmax=100)
      OK       <- predict(object=ok_gstat, newdata=df_hidded_AQMS)
      
      
      OK_other_obs           <- data.frame(data=OK$var1.pred)
      colnames(OK_other_obs) <- station
      OK_other_obs$date      <- stations_to_interp$date[j]
      
      OK_other_obs_df <-  plyr::rbind.fill(as.data.frame(OK_other_obs_df), OK_other_obs)
      
    } else{
      #print(paste0("There is no observations in: ", stations_to_interp$date[j]))
      OK_other_obs           <- data.frame(data=NA)
      colnames(OK_other_obs) <- station
      OK_other_obs$date      <- stations_to_interp$date[j]
      
      OK_other_obs_df        <-  plyr::rbind.fill(as.data.frame(OK_other_obs_df), OK_other_obs)
      c <- c+1
    }
    
  }
  print(c)
  
  # Select station data
  single_station_data <- data.frame(date            = OK_other_obs_df$date,
                                    obs             = observations[,station],
                                    model           = df_all_model_raw_at_AQMS[,station], 
                                    day_of_year     = yday(OK_other_obs_df$date),
                                    day_of_week     = wday(OK_other_obs_df$date),
                                    other_obs       = OK_other_obs_df[,station])
  #estacions interpolat a l'estació
  
  single_station_data                     <- na.omit(single_station_data)
  
  # Split train/test fractions 
  train.fraction  <- 0.90
  smp_size        <- floor(train.fraction * nrow(single_station_data)) #te redondea el número i ya
  
  #seq_len et cre una seqüència de números nrow=47 seq_len de l'1 al 47
  #sample agafa un subconjunt aleatòri d'elements d'un vector, en aquest cas agafa el 90% (size)
  set.seed(21) #li he afegit, ns si te sentit que a cada estacio agafi el mateix train_id ---> Preguntar JAN
  train_ind     <- sample(seq_len(nrow(single_station_data)), size = smp_size) #42 valors aleatòris (90% de 48 h)
  data_training <- single_station_data[train_ind,]
  data_test     <- single_station_data[-train_ind,]
  
  #test_ind <- 140 : (140 + 48) #??? Osigui tens 48 valors pero que comencen en el 140??
  #data_training = single_station_data[-test_ind,] #?? Suposo que te a veure amb el 90% este i agafa la part central, ni idea
  #data_test = single_station_data[test_ind,] #??
  
  # Preliminary fit
  #va rapit amb pm2,5, pot ser guay fer bucles per justificar millor pk agafem els numeros que agafem (ma o menos)
  n.trees = 5000 #realment s'ha de provar i escollir l'òptim (si dona temps bucle amb 3-4 opcions per comparar) [la que dongui menor mse]
  #Vx_formula <- "obs ~ model + hour_of_the_day + day_of_year + day_of_week"
  Vx_formula <- "obs ~ model + day_of_year + day_of_week+ other_obs"
  gbm1 <- gbm(as.formula(Vx_formula),
              data = data_training,
              distribution = "gaussian",
              n.trees = n.trees,
              interaction.depth = 1, #amb 1 hauria de funcionar, no solem posar més de 10
              shrinkage = 0.01, #learning rate ni molt petit pk anirà lent ni molt gran pk te pasas [ns si bucle amb 3 proves]
              cv.folds = 5, #cross validation
              n.cores = NULL, # will use all cores by default
              verbose = FALSE)  
  gbm.fit<- gbm1
  
  # print results
  print(gbm.fit)
  
  # plot loss function as a result of n trees added to the ensemble
  gbm.perf(gbm.fit, method = "cv")
  
  # get MSE 
  sqrt(min(gbm.fit$cv.error))
  
  # Variable importance
  gbm.fit.final <- gbm.fit
  par(mar = c(5, 8, 1, 1))
  summary(
    gbm.fit.final, 
    cBars = 10,
    method = relative.influence, # also can use permutation.test.gbm
    las = 2)
  
#Once you have decided on a final model you will likely want to use the model to predict on new observations. 
  res_training <- predict( gbm1 , newdata=data_training)
  #res_test     <- predict( gbm1 , newdata=data_test[,c("model","hour_of_the_day","day_of_year","day_of_week")])
  res_test     <- predict( gbm1 , newdata=data_test[,c("model","day_of_year","day_of_week","other_obs")])
  
  Test_results <- data.frame(date = data_test[,'date'], Test_results=res_test )
  Test_results_station[Test_results_station$date%in%Test_results$date ,"Test_results"] <- Test_results[,"Test_results"]
  
  Obs_Test <- data.frame(date = data_test[,'date'], Obs_Test=data_test$obs )
  Obs_Test_station[Obs_Test_station$date%in%Obs_Test$date ,"Obs_Test"] <- Obs_Test[,"Obs_Test"]
  
  Obs_Train <- data.frame(date = data_training[,'date'], Obs_Train=data_training$obs )
  Obs_Train_station[Obs_Train_station$date%in%Obs_Train$date ,"Obs_Train"] <- Obs_Train[,"Obs_Train"]
  
  data_single     <- data.frame(date                    = df_all_obs$date,
                                Obs_Train_station       = Obs_Train_station$Obs_Train,
                                Obs_Test_station        = Obs_Test_station$Obs_Test,
                                Test_results_station    = Test_results_station$Test_results,
                                model                   = df_all_model_raw_at_AQMS[,station])
  
  #mm <- "January"
  # TIME_PLOT
  #png(file = paste0(save_time_series,local_stations$ref_obs[l],"_", local_stations$station_name[l],"_",mm,".png"), width = 800, height = 400)
  openair::timePlot(data_single,
                    pollutant = c("Obs_Train_station","Obs_Test_station","Test_results_station","model"),
                    group = TRUE,
                    title=station,
                    plot.type= "h",
                    y.relation = "free",
                    ylab="PM2,5 (ug/m3)",
                    cols = c("black", "blue", "red", "green"),
                    lty = c("dotted","solid","solid","dashed"),
                    lwd = c(3,2,2,2), 
                    name.pol = c("Obs Train station","Obs Test station", "Test results station", "Model"))
  dev.off()
  
  #analisi de la cross validation
  test_performace     <- postResample(pred = res_test, obs = data_test$obs)
  training_performace <- postResample(pred = res_training, obs = data_training$obs)
  
}

Summary_stats <- openair::modStats(data_single, obs = "Obs_Test_station", mod = "Test_results_station")
print(Summary_stats)
summary(data_single)
